/*******************************************************************************
  Copyright(c) 2017 Jasem Mutlaq. All rights reserved.

  List of INDI Stanadrd Properties

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
 License version 2 as published by the Free Software Foundation.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.

 You should have received a copy of the GNU Library General Public License
 along with this library; see the file COPYING.LIB.  If not, write to
 the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 Boston, MA 02110-1301, USA.
*******************************************************************************/

#include "indistandardproperty.h"

namespace INDI
{
namespace SP
{
const char *CONNECTION = "CONNECTION";
const char *DEVICE_PORT = "DEVICE_PORT";
const char *DEVICE_AUTO_SEARCH = "DEVICE_AUTO_SEARCH";
const char *DEVICE_LAN_SEARCH = "DEVICE_LAN_SEARCH";
const char *DEVICE_BAUD_RATE = "DEVICE_BAUD_RATE";
const char *DEVICE_ADDRESS = "DEVICE_ADDRESS";
} // namespace SP
} // namespace INDI
